@@ -22,10 +22,15 @@ case ${os} in
       7)
         DNF=yum
         ;;
-      *)
+      8)
         DNF="dnf --enablerepo=powertools"
         ${DNF} module -y disable postgresql
         ;;
+      *)
+        DNF="dnf --enablerepo=crb"
+        ${DNF} install -y \
+               https://apache.jfrog.io/artifactory/arrow/${os}/${major_version}/apache-arrow-release-latest.rpm
+        ;;
     esac
 
     ${DNF} install -y \
@@ -45,9 +50,9 @@ echo "::group::Install built packages"
 
 packages_dir=/host/repositories/${os}/${major_version}/x86_64/Packages
 
-postgresql_version=$(basename $(ls ${packages_dir}/*-pgroonga-*.rpm | head -n1) | \
-                       grep -E -o '[0-9.]+' |
-                       head -n1)
+pgroonga_package=$(basename $(ls ${packages_dir}/*-pgroonga-*.rpm | head -n1) | \
+                     sed -e 's/-pgroonga-.*$/-pgroonga/g')
+postgresql_version=$(echo ${pgroonga_package} | grep -E -o '[0-9.]+')
 case ${os} in
   amazon-linux)
     amazon-linux-extras install -y postgresql${postgresql_version}
@@ -141,8 +146,6 @@ case "${os}" in
     rm sql/function/match-positions-character/one-keyword.sql
     ;;
 esac
-# TODO: Remove this after Groonga 13.0.1 is released.
-rm sql/function/highlight-html/term-search.sql
 ruby /host/test/prepare.rb > schedule
 PG_REGRESS_DIFF_OPTS="-u"
 if diff --help | grep -q color; then
@@ -174,6 +177,19 @@ if [ ${pg_regress_status} -ne 0 ]; then
   exit ${pg_regress_status}
 fi
 
+
+echo "::group::Upgrade"
+
+${DNF} remove -y ${pgroonga_package}
+${DNF} install -y ${pgroonga_package}
+createdb upgrade
+psql upgrade -c 'CREATE EXTENSION pgroonga'
+${DNF} install -y ${packages_dir}/*.rpm
+psql upgrade -c 'ALTER EXTENSION pgroonga UPDATE'
+
+echo "::endgroup::"
+
+
 echo "::group::Postpare"
 
 sudo -u postgres -H \
