@@ -11,6 +11,8 @@
 #ifdef PGRN_SUPPORT_TABLEAM
 #	include <access/tableam.h>
 #endif
+#include <catalog/pg_type.h>
+#include <funcapi.h>
 #include <miscadmin.h>
 
 static bool PGrnWALEnabled = false;
@@ -63,6 +65,7 @@ PGrnWALSetMaxSize(size_t size)
 
 PGDLLEXPORT PG_FUNCTION_INFO_V1(pgroonga_wal_apply_index);
 PGDLLEXPORT PG_FUNCTION_INFO_V1(pgroonga_wal_apply_all);
+PGDLLEXPORT PG_FUNCTION_INFO_V1(pgroonga_wal_status);
 PGDLLEXPORT PG_FUNCTION_INFO_V1(pgroonga_wal_truncate_index);
 PGDLLEXPORT PG_FUNCTION_INFO_V1(pgroonga_wal_truncate_all);
 
@@ -209,15 +212,6 @@ msgpack_pack_grn_obj(msgpack_packer *packer, grn_obj *object)
 #define PGRN_WAL_META_PAGE_BLOCK_NUMBER 0
 
 #ifdef PGRN_SUPPORT_WAL
-static LOCKMODE
-PGrnWALLockMode(void)
-{
-	if (RecoveryInProgress())
-		return RowExclusiveLock;
-	else
-		return ShareUpdateExclusiveLock;
-}
-
 static Buffer
 PGrnWALReadLockedBuffer(Relation index,
 						BlockNumber blockNumber,
@@ -236,6 +230,30 @@ PGrnWALReadLockedBuffer(Relation index,
 	return buffer;
 }
 
+static char *
+PGrnWALPageGetData(Page page)
+{
+	return PageGetContents(page);
+}
+
+static size_t
+PGrnWALPageGetFreeSize(Page page)
+{
+	PageHeader pageHeader;
+
+	pageHeader = (PageHeader)page;
+	return pageHeader->pd_upper - pageHeader->pd_lower;
+}
+
+static LocationIndex
+PGrnWALPageGetLastOffset(Page page)
+{
+	PageHeader pageHeader;
+
+	pageHeader = (PageHeader)page;
+	return pageHeader->pd_lower - SizeOfPageHeaderData;
+}
+
 static void
 PGrnWALDataInitBuffers(PGrnWALData *data)
 {
@@ -308,9 +326,29 @@ PGrnWALDataInitCurrent(PGrnWALData *data)
 }
 
 static void
-PGrnWALDataRestart(PGrnWALData *data)
+PGrnWALDataFinish(PGrnWALData *data)
 {
 	GenericXLogFinish(data->state);
+	if (data->current.page)
+	{
+		PGrnIndexStatusSetWALAppliedPosition(
+			data->index,
+			BufferGetBlockNumber(data->current.buffer),
+			PGrnWALPageGetLastOffset(data->current.page));
+	}
+	else
+	{
+		PGrnIndexStatusSetWALAppliedPosition(
+			data->index,
+			data->meta.pageSpecial->next,
+			0);
+	}
+}
+
+static void
+PGrnWALDataRestart(PGrnWALData *data)
+{
+	PGrnWALDataFinish(data);
 
 	PGrnWALDataReleaseBuffers(data);
 
@@ -320,30 +358,6 @@ PGrnWALDataRestart(PGrnWALData *data)
 	PGrnWALDataInitCurrent(data);
 }
 
-static char *
-PGrnWALPageGetData(Page page)
-{
-	return PageGetContents(page);
-}
-
-static size_t
-PGrnWALPageGetFreeSize(Page page)
-{
-	PageHeader pageHeader;
-
-	pageHeader = (PageHeader)page;
-	return pageHeader->pd_upper - pageHeader->pd_lower;
-}
-
-static LocationIndex
-PGrnWALPageGetLastOffset(Page page)
-{
-	PageHeader pageHeader;
-
-	pageHeader = (PageHeader)page;
-	return pageHeader->pd_lower - SizeOfPageHeaderData;
-}
-
 static void
 PGrnWALPageAppend(Page page, const char *data, size_t dataSize)
 {
@@ -454,10 +468,6 @@ PGrnWALPageWriter(void *userData,
 		if (rest <= freeSize)
 		{
 			PGrnWALPageAppend(data->current.page, buffer, rest);
-			PGrnIndexStatusSetWALAppliedPosition(
-				data->index,
-				BufferGetBlockNumber(data->current.buffer),
-				PGrnWALPageGetLastOffset(data->current.page));
 			written += rest;
 		}
 		else
@@ -480,6 +490,36 @@ PGrnWALDataInitMessagePack(PGrnWALData *data)
 {
 	msgpack_packer_init(&(data->packer), data, PGrnWALPageWriter);
 }
+
+static LOCKMODE
+PGrnWALLockMode(void)
+{
+	if (RecoveryInProgress())
+		return RowExclusiveLock;
+	else
+		return ShareUpdateExclusiveLock;
+}
+
+static BlockNumber
+PGrnWALLockBlockNumber(void)
+{
+	/* We can use any block number for this. We just want an index
+	 * level lock but we can't use LockRelation(index) because it
+	 * conflicts with REINDEX INDEX CONCURRENTLY. */
+	return 0;
+}
+
+static void
+PGrnWALLock(Relation index)
+{
+	LockPage(index, PGrnWALLockBlockNumber(), PGrnWALLockMode());
+}
+
+static void
+PGrnWALUnlock(Relation index)
+{
+	UnlockPage(index, PGrnWALLockBlockNumber(), PGrnWALLockMode());
+}
 #endif
 
 PGrnWALData *
@@ -494,7 +534,7 @@ PGrnWALStart(Relation index)
 	if (!RelationIsValid(index))
 		return NULL;
 
-	LockRelation(index, PGrnWALLockMode());
+	PGrnWALLock(index);
 
 	data = palloc(sizeof(PGrnWALData));
 
@@ -520,11 +560,11 @@ PGrnWALFinish(PGrnWALData *data)
 	if (!data)
 		return;
 
-	GenericXLogFinish(data->state);
+	PGrnWALDataFinish(data);
 
 	PGrnWALDataReleaseBuffers(data);
 
-	UnlockRelation(data->index, PGrnWALLockMode());
+	PGrnWALUnlock(data->index);
 
 	pfree(data);
 #endif
@@ -548,7 +588,7 @@ PGrnWALAbort(PGrnWALData *data)
 	{
 		PGrnWALDataReleaseBuffers(data);
 
-		UnlockRelation(data->index, PGrnWALLockMode());
+		PGrnWALUnlock(data->index);
 	}
 
 	pfree(data);
@@ -2150,13 +2190,13 @@ PGrnWALApply(Relation index)
 	if (!PGrnWALApplyNeeded(&data))
 		return 0;
 
-	LockRelation(index, PGrnWALLockMode());
+	PGrnWALLock(index);
 	PGrnIndexStatusGetWALAppliedPosition(data.index,
 										 &(data.current.block),
 										 &(data.current.offset));
 	data.sources = NULL;
 	nAppliedOperations = PGrnWALApplyConsume(&data);
-	UnlockRelation(index, PGrnWALLockMode());
+	PGrnWALUnlock(index);
 #endif
 	return nAppliedOperations;
 }
@@ -2292,6 +2332,152 @@ pgroonga_wal_apply_all(PG_FUNCTION_ARGS)
 	PG_RETURN_INT64(nAppliedOperations);
 }
 
+typedef struct {
+	Relation indexes;
+	PGrnTableScanDesc scan;
+	TupleDesc desc;
+} PGrnWALStatusData;
+
+/**
+ * pgroonga_wal_status() : SETOF RECORD
+ */
+Datum
+pgroonga_wal_status(PG_FUNCTION_ARGS)
+{
+	FuncCallContext *context;
+#ifdef PGRN_SUPPORT_WAL
+	LOCKMODE lock = AccessShareLock;
+	PGrnWALStatusData *data;
+	HeapTuple indexTuple;
+
+	if (SRF_IS_FIRSTCALL())
+	{
+		MemoryContext oldContext;
+		context = SRF_FIRSTCALL_INIT();
+		oldContext = MemoryContextSwitchTo(context->multi_call_memory_ctx);
+		PG_TRY();
+		{
+			const int nAttributes = 8;
+			data = palloc(sizeof(PGrnWALStatusData));
+			data->indexes = pgrn_table_open(IndexRelationId, lock);
+			data->scan = pgrn_table_beginscan_catalog(data->indexes, 0, NULL);
+			data->desc = PGrnCreateTemplateTupleDesc(nAttributes);
+			TupleDescInitEntry(data->desc, 1, "name", TEXTOID, -1, 0);
+			TupleDescInitEntry(data->desc, 2, "oid", OIDOID, -1, 0);
+			TupleDescInitEntry(data->desc, 3, "current_block", INT8OID, -1, 0);
+			TupleDescInitEntry(data->desc, 4, "current_offset", INT8OID, -1, 0);
+			TupleDescInitEntry(data->desc, 5, "current_size", INT8OID, -1, 0);
+			TupleDescInitEntry(data->desc, 6, "last_block", INT8OID, -1, 0);
+			TupleDescInitEntry(data->desc, 7, "last_offset", INT8OID, -1, 0);
+			TupleDescInitEntry(data->desc, 8, "last_size", INT8OID, -1, 0);
+			BlessTupleDesc(data->desc);
+			context->user_fctx = data;
+		}
+		PG_CATCH();
+		{
+			MemoryContextSwitchTo(oldContext);
+			PG_RE_THROW();
+		}
+		PG_END_TRY();
+		MemoryContextSwitchTo(oldContext);
+		context->tuple_desc = data->desc;
+	}
+
+	context = SRF_PERCALL_SETUP();
+	data = context->user_fctx;
+
+	while ((indexTuple = heap_getnext(data->scan, ForwardScanDirection)))
+	{
+		Form_pg_index indexForm = (Form_pg_index) GETSTRUCT(indexTuple);
+		Relation index;
+		BlockNumber currentBlock;
+		LocationIndex currentOffset;
+		Datum *values;
+		bool *nulls;
+		int i = 0;
+
+		if (!pgrn_pg_class_ownercheck(indexForm->indexrelid, GetUserId()))
+			continue;
+
+		index = RelationIdGetRelation(indexForm->indexrelid);
+		if (!PGrnIndexIsPGroonga(index))
+		{
+			RelationClose(index);
+			continue;
+		}
+		if (PGRN_RELKIND_HAS_PARTITIONS(index->rd_rel->relkind))
+		{
+			RelationClose(index);
+			continue;
+		}
+
+		values = palloc(sizeof(Datum) * data->desc->natts);
+		/* All values are not NULL */
+		nulls = palloc0(sizeof(bool) * data->desc->natts);
+		values[i++] = PointerGetDatum(cstring_to_text(RelationGetRelationName(index)));
+		values[i++] = ObjectIdGetDatum(RelationGetRelid(index));
+		PGrnIndexStatusGetWALAppliedPosition(index, &currentBlock, &currentOffset);
+		values[i++] = Int64GetDatum(currentBlock);
+		values[i++] = Int64GetDatum(currentOffset);
+		values[i++] = Int64GetDatum(currentBlock * BLCKSZ + currentOffset);
+		if (PGrnWALEnabled && RelationGetNumberOfBlocks(index) > 0)
+		{
+			Buffer metaBuffer =
+				PGrnWALReadLockedBuffer(index,
+										PGRN_WAL_META_PAGE_BLOCK_NUMBER,
+										BUFFER_LOCK_SHARE);
+			Page metaPage = BufferGetPage(metaBuffer);
+			PGrnWALMetaPageSpecial *metaPageSpecial =
+				(PGrnWALMetaPageSpecial *) PageGetSpecialPointer(metaPage);
+			BlockNumber lastBlock = metaPageSpecial->next;
+			LocationIndex lastOffset = 0;
+			if (lastBlock < RelationGetNumberOfBlocks(index))
+			{
+				Buffer lastBuffer =
+					PGrnWALReadLockedBuffer(index,
+											metaPageSpecial->next,
+											BUFFER_LOCK_SHARE);
+				Page lastPage = BufferGetPage(lastBuffer);
+				lastOffset = PGrnWALPageGetLastOffset(lastPage);
+				UnlockReleaseBuffer(lastBuffer);
+			}
+			values[i++] = Int64GetDatum(lastBlock);
+			values[i++] = Int64GetDatum(lastOffset);
+			values[i++] = Int64GetDatum(lastBlock * BLCKSZ + lastOffset);
+			UnlockReleaseBuffer(metaBuffer);
+		}
+		else
+		{
+			values[i++] = Int64GetDatum(0);
+			values[i++] = Int64GetDatum(0);
+			values[i++] = Int64GetDatum(0);
+		}
+		RelationClose(index);
+
+		{
+			HeapTuple tuple = heap_form_tuple(data->desc, values, nulls);
+			SRF_RETURN_NEXT(context, HeapTupleGetDatum(tuple));
+		}
+	}
+
+	heap_endscan(data->scan);
+	pgrn_table_close(data->indexes, lock);
+#else
+	if (SRF_IS_FIRSTCALL())
+	{
+		context = SRF_FIRSTCALL_INIT();
+	}
+	context = SRF_PERCALL_SETUP();
+	{
+		const char *tag = "[wal][status]";
+		PGrnCheckRC(GRN_FUNCTION_NOT_IMPLEMENTED,
+					"%s not supported",
+					tag);
+	}
+#endif
+	SRF_RETURN_DONE(context);
+}
+
 #ifdef PGRN_SUPPORT_WAL
 static int64_t
 PGrnWALTruncate(Relation index)
@@ -2303,11 +2489,11 @@ PGrnWALTruncate(Relation index)
 	uint32_t nProcessingBuffers = 0;
 	GenericXLogState *state;
 
-	LockRelation(index, PGrnWALLockMode());
+	PGrnWALLock(index);
 	nBlocks = RelationGetNumberOfBlocks(index);
 	if (nBlocks == 0)
 	{
-		UnlockRelation(index, PGrnWALLockMode());
+		PGrnWALUnlock(index);
 		return 0;
 	}
 
@@ -2368,7 +2554,7 @@ PGrnWALTruncate(Relation index)
 										 PGRN_WAL_META_PAGE_BLOCK_NUMBER + 1,
 										 0);
 
-	UnlockRelation(index, PGrnWALLockMode());
+	PGrnWALUnlock(index);
 
 	return nTruncatedBlocks;
 }
